@@ -5,7 +5,13 @@ import tkinter as tk
 
 import contextlib
 import sys
-import multiprocessing
+
+import threading
+
+global global_vars
+global_vars = {}
+global local_vars
+local_vars = {}
 
 print(chr(27) + "[2J")
 
@@ -58,13 +64,13 @@ class TextIO:
         return self.user_input
 
 
-class Cell:
+class Cell():
 
     def __init__(self,
                  name_,
-                 graph_,
                  content_type_="python",
-                 content_=" "):
+                 content_=" ",
+                 stdout="internal"):
         """
         :param name_: The cell's name
         :param graph_: The cell's parent graph
@@ -74,9 +80,11 @@ class Cell:
         self.name = name_
         self.content_type = content_type_
         self.content = content_
-        self.graph = graph_
+        self.stdout = stdout
+        self.output = ""
 
-        self.thread = None
+    def get_copy(self):
+        return Cell(self.name, self.content_type, self.content, self.stdout)
 
     @contextlib.contextmanager
     def stdoutIO(self, stdout=None):
@@ -87,33 +95,34 @@ class Cell:
         yield stdout
         sys.stdout = old
 
-    def execute(self, stdout):
-        self.thread = multiprocessing.Process(target=self.execute_in_thread, args=(stdout,))
-        self.thread.daemon = True
-        self.thread.start()
+    def get_vars(self):
+        return self.self_globals, self.self_locals
 
-    def kill_cell(self):
-        self.thread.terminate()
-
-    def execute_in_thread(self, stdout):
+    def execute(self):
         # Execute this cell's content
-        if stdout == "internal":
+
+        gcopy = global_vars.copy()
+        lcopy = local_vars.copy()
+
+        if self.stdout == "internal":
             try:
-                exec(self.content, self.graph.exec_globals)
+                exec(self.content, gcopy, lcopy)
             except Exception as e:
                 print("Exception occurred in cell " + self.name)
                 print(e)
-        elif stdout == "external":
+        elif self.stdout == "external":
             try:
                 with self.stdoutIO() as s:
-                    exec(self.content, self.graph.exec_globals)
-                    return s.getvalue()
+                    exec(self.content, gcopy, lcopy)
+                    self.output = s.getvalue()
             except Exception as e:
                 print("Exception occurred in cell " + self.name)
                 print(e)
         else:
-            print("stdout setting \"" + stdout + "\" not recognized. Please use internal/external.")
-            return
+            print("stdout setting \"" + self.stdout + "\" not recognized. Please use internal/external.")
+
+        global_vars.update(gcopy)
+        local_vars.update(lcopy)
 
     def __str__(self):
         return self.name + "\n\n" + "```\n" + self.content + "```\n"
@@ -126,8 +135,11 @@ class Graph:
         self.graph = nx.DiGraph()
         # Dict to keep track of cell names vs networkx node names
         self.names_to_indeces = {}
-        # Dictionary for variables created by cells
-        self.exec_globals = {}
+        # Dictionaries for variables created by cells
+        global global_vars
+        global_vars = {}
+        global local_vars
+        local_vars = {}
         # TextIO object
         self.ti = TextIO()
 
@@ -135,9 +147,9 @@ class Graph:
         return {idx: moniker for moniker, idx in
                 zip(list(self.names_to_indeces.keys()), list(self.names_to_indeces.values()))}
 
-    def reset_runtime(self):
-        # Erase all runtime variables
-        self.exec_globals = {}
+    def update_vars(self, new_globals, new_locals):
+        self.exec_globals.update(new_globals)
+        self.exec_locals.update(new_locals)
 
     def name_to_idx(self, cell_name):
         """
@@ -158,7 +170,7 @@ class Graph:
         """
         cells = list(nx.get_node_attributes(self.graph, 'data').values())
 
-        if cell_index:
+        if cell_index is not None:
             output = cells[cell_index]
             return output
 
@@ -211,8 +223,12 @@ class Graph:
         self.graph.remove_edge(idx1, idx2)
 
     def swap_cells(self, name1, name2):
-        cell1 = Cell(name2, self, self.get_cell(name2).content_type, self.get_cell(name2).content)
-        cell2 = Cell(name1, self, self.get_cell(name1).content_type, self.get_cell(name1).content)
+
+        old_cell1 = self.get_cell(name1)
+        old_cell2 = self.get_cell(name2)
+
+        cell1 = Cell(name2, old_cell2.content_type, old_cell2.content)
+        cell2 = Cell(name1, old_cell1.content_type, old_cell1.content)
 
         idx1 = self.name_to_idx(name1)
         idx2 = self.name_to_idx(name2)
@@ -231,8 +247,17 @@ class Graph:
             print("To merge, cells must be adjacent")
             return
 
+        c1 = self.get_cell("", idx1)
+        c2 = self.get_cell("", idx2)
+
+        glb = c1.self_globals
+        lcl = c1.self_locals
+
+        glb.update(c2.self_globals)
+        lcl.update(c2.self_locals)
+
         # make new cell
-        new_cell = Cell(name_=new_name, graph_=self, content_type_=self.get_cell("", idx1).content_type)
+        new_cell = Cell(new_name, content_type_=self.get_cell("", idx1).content_type)
         new_content = self.get_cell("", idx1).content + "\n# merge point\n" + self.get_cell("", idx2).content
         new_cell.content = new_content
         self.names_to_indeces[new_name] = self.names_to_indeces[self.get_cell("", idx1).name]
@@ -296,17 +321,66 @@ class Graph:
 
         return in_edges, out_edges
 
-    def execute_linear_list_of_cells(self, cells_list=None, stdout="internal", output_filename="stdout.txt"):
-        if not cells_list:
-            cells_list = self.get_all_cells_edges()[0]
+    def execute_linear_list_of_cells(self, cells_list, stdout="internal", output_filename="stdout.txt"):
+        std_file_out = ""
 
         for cell_name in cells_list:
             cell = self.get_cell(cell_name)
+
+            cell.stdout = stdout
+
+            p = threading.Thread(target=cell.execute)
+
             if cell.content_type == "python":
-                cell_output = cell.execute(stdout)
-                if stdout == "external":
-                    with open(output_filename, 'w') as txt:
-                        txt.write(cell_output)
+                p.start()
+                p.join()
+
+            std_file_out += cell.output
+
+        if stdout == "external":
+            with open(output_filename, 'w') as txt:
+                txt.write(std_file_out)
+
+    def bfs_traversal_execute(self, stdout="internal", output_filename="stdout.txt"):
+        std_file_out = ""
+
+        root_cell = self.get_cell("", 0)
+        root_cell.stdout = stdout
+
+        root = threading.Thread(target=root_cell.execute)
+
+        root.start()
+        root.join()
+
+        std_file_out += root_cell.output
+
+        neighbors = self.graph.neighbors(0)
+
+        while neighbors:
+            new_neighbors = []
+            processes = []
+            for n in neighbors:
+                neighbor_cell = self.get_cell(self.get_lookup_table()[n])
+                neighbor_cell.stdout = stdout
+
+                neighbor = threading.Thread(target=neighbor_cell.execute)
+
+                neighbor.start()
+                processes.append(neighbor)
+
+                new_neighbors.extend(self.graph.neighbors(n))
+
+            [proc.join() for proc in processes]
+
+            for n in neighbors:
+                neighbor = self.get_cell(self.get_lookup_table()[n])
+                std_file_out += neighbor.output
+
+            neighbors = new_neighbors
+
+        if stdout == "external":
+            with open(output_filename, 'w') as txt:
+                txt.write(std_file_out)
 
     def save_graph(self, filename):
         txtout = ""
@@ -335,7 +409,6 @@ class Graph:
             file.write(txtout)
 
 
-# noinspection PyBroadException
 class Interpreter:
 
     def __init__(self):
@@ -445,7 +518,7 @@ class Interpreter:
                 ti = TextIO()
                 content = ti.text_input().strip()
 
-        self.graph.add_cell(Cell(name, self.graph, content_type, content))
+        self.graph.add_cell(Cell(name, content_type, content))
 
     def edit_cell(self, command):
         """
@@ -539,19 +612,19 @@ class Interpreter:
         """
         :param command: command to be executed
         """
-        if len(command) > 1:
+        if ">>" in command:
+            cells_list = command[1:-2]
+            self.stdout_filename = command[-1]
+        else:
+            cells_list = command[1:]
+        if len(cells_list) >= 1:
             try:
-                if ">>" in command:
-                    cells_list = command[1:-2]
-                    self.stdout_filename = command[-1]
-                else:
-                    cells_list = command[1:]
                 self.graph.execute_linear_list_of_cells(cells_list, self.stdout, self.stdout_filename)
             except Exception as e:
                 print("There was an error executing one of the cells")
                 print(e)
         else:
-            self.graph.execute_linear_list_of_cells(None, self.stdout)
+            self.graph.bfs_traversal_execute(self.stdout, self.stdout_filename)
 
     def display(self, command):
         """
@@ -598,7 +671,10 @@ class Interpreter:
 
     def reset_runtime(self):
         # Delete all runtime variables
-        self.graph.reset_runtime()
+        global global_vars
+        global_vars = {}
+        global local_vars
+        local_vars = {}
 
     def reset_graph(self):
         confirm = input("Are you sure you want to reset the graph? This will delete all nodes and variables. (y/n) ")
